/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 2011 OpenFOAM Foundation
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

\*---------------------------------------------------------------------------*/

#include "solidParticleCloud.H"

// * * * * * * * * * * * * * * Static Data Members * * * * * * * * * * * * * //

namespace Foam
{
    defineTemplateTypeNameAndDebug(Cloud<solidParticle>, 0);
}

// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

bool Foam::solidParticle::move
(
    trackingData& td,
    const scalar trackTime
)
{
    td.switchProcessor = false;
    td.keepParticle = true;

    const polyBoundaryMesh& pbMesh = mesh_.boundaryMesh();

    scalar tEnd = (1.0 - stepFraction())*trackTime;
    scalar dtMax = tEnd;

    while (td.keepParticle && !td.switchProcessor && tEnd > SMALL)
    {
        if (debug)
        {
            Info<< "Time = " << mesh_.time().timeName()
                << " trackTime = " << trackTime
                << " tEnd = " << tEnd
                << " steptFraction() = " << stepFraction() << endl;
        }

        // set the lagrangian time-step
        scalar dt = min(dtMax, tEnd);

        // remember which cell the parcel is in
        // since this will change if a face is hit
        label cellI = cell();

        dt *= trackToFace(position() + dt*U_, td);

        tEnd -= dt;
        stepFraction() = 1.0 - tEnd/trackTime;

        cellPointWeight cpw(mesh_, position(), cellI, face());
        scalar rhoc = td.rhoInterp().interpolate(cpw);
        vector Uc = td.UInterp().interpolate(cpw);
        scalar nuc = td.nuInterp().interpolate(cpw);
        scalar pc  = td.pInterp().interpolate(cpw); 

        scalar rhop = td.cloud().rhop();
        scalar magUr = mag(Uc - U_);

        scalar ReFunc = 1.0;
        scalar Re = magUr*d_/nuc;

        if (Re > 0.01)
        {
            ReFunc += 0.15*pow(Re, 0.687);
        }

        scalar Dc = (24.0*nuc/d_)*ReFunc*(3.0/4.0)*(rhoc/(d_*rhop));
        
        scalar m = rhop*(4.0/3.0)*constant::mathematical::pi*pow(d_/2.0,3);
        vector oldMom = U_*m;

        U_ = (U_ + dt*(Dc*Uc + (1.0 - rhoc/rhop)*td.g()))/(1.0 + dt*Dc);
        
        vector newMom = U_*m;
        td.cloud().smom()[cellI] += newMom-oldMom;
        
        /* solving the Rayleigh-Plesset with the Rosenbrock method */
 /* ======================================================= */
	scalar hrk = 1.0E-10;
	scalar h   = 0.5*hrk;
	scalar dD  = 1.0/(2.0+sqrt(2.0));
	scalar time_rk = 0;
	scalar sigma = td.cloud().sigma();
	scalar pV = td.cloud().pSat();
	
	rhoc = 998.85;
	nuc  = 6.93305e-04;
	
	scalar substeps = 0;
			
	while (time_rk < trackTime)
	  {
	    
	    if (time_rk+h > trackTime)
	      {
		h = trackTime-time_rk;
	      }
    
		// f 
	scalar F0 = Rdt_;
	scalar F1 = -1.5*(Rdt_*Rdt_/R_)+((pB_-pc)/(rhoc*R_))-4.0*nuc*(Rdt_/(R_*R_))-((2.0*sigma)/(rhoc*R_*R_));

	// f0
	scalar F00 = F0;
	scalar F01 = F1;

	// the jacobian dF/dX
	scalar J00 = 0.0;
	scalar J01 = 1.0;
	scalar J10 = 1.5*((Rdt_*Rdt_)/(R_*R_))-((pB_-pc)/(rhoc*R_*R_))+
	  ((8.0*nuc*Rdt_)/(R_*R_*R_))+((4.0*sigma)/(rhoc*R_*R_*R_));
	scalar J11 = -3.0*Rdt_/R_-4.0*(nuc/(R_*R_));
	
	// W = I - h dD J
	scalar W00 = 1.0 - h*dD*J00;
	scalar W01 = 0.0 - h*dD*J01;
	scalar W10 = 0.0 - h*dD*J10;
	scalar W11 = 1.0 - h*dD*J11;
	
	// inv W
	scalar ff = 1.0/(W00*W11-W01*W10);
	scalar invW00 =  ff*W11;
	scalar invW01 = -ff*W01;
	scalar invW10 = -ff*W10;
	scalar invW11 =  ff*W00;
	
	// the time derivative of function, dF/dt
	// estimated numerically
	scalar T0 = (F0-F0Old_)/h;
	scalar T1 = (F1-F1Old_)/h;

	// k1
	scalar tmp0 = F00+h*dD*T0;
	scalar tmp1 = F01+h*dD*T1;
	scalar k10 = invW00*tmp0+invW01*tmp1;
	scalar k11 = invW10*tmp0+invW11*tmp1;

	// y_n + 0.5 h k1
	scalar R1   = R_+0.5*h*k10;
	scalar Rdt1 = Rdt_+0.5*h*k11;

	// f1
	scalar F10 = Rdt1;
	scalar F11 = -1.5*(Rdt1*Rdt1/R1)+((pB_-pc)/(rhoc*R1))-4.0*nuc*(Rdt1/(R1*R1))-((2.0*sigma)/(rhoc*R1*R1));

	// k2
	tmp0       = F10-k10;
	tmp1       = F11-k11;
	scalar k20 = invW00*tmp0+invW01*tmp1+k10;
	scalar k21 = invW10*tmp0+invW11*tmp1+k11;   

	// y_n+1
	scalar R2   = R_+h*k20;
	scalar Rdt2 = Rdt_+h*k21;

	
	// f2
	scalar F20 = Rdt2;
	scalar F21 = -1.5*(Rdt2*Rdt2/R2)+((pB_-pc)/(rhoc*R2))-4.0*nuc*(Rdt2/(R2*R2))-((2.0*sigma)/(rhoc*R2*R2));
	
	// k3	
	scalar b   = 6.0 + sqrt(2.0);
	tmp0       = F20-b*(k20-F10)-2.0*(k10-F00)+h*dD*T0;
	tmp1       = F21-b*(k21-F11)-2.0*(k11-F01)+h*dD*T1;
	scalar k30 = invW00*tmp0+invW01*tmp1;
	scalar k31 = invW10*tmp0+invW11*tmp1;
	// error
	scalar err0 = fabs((h/6.0)*(k10-2.0*k20+k30));
	scalar err1 = fabs((h/6.0)*(k11-2.0*k21+k31));
	scalar err  = max(err0,err1);
	
	time_rk  += h;
        substeps += 1;
	
	scalar absTol = 1.0E-5;

	
	if (err > 1.0E-08)
	  {
	    h = min(0.5*hrk, 0.5*h);
	  }
	else
	  {
	    h = max(1.5*hrk, 1.5*h);
	  }
	// update bubble state
	F0Old_ = F0;
	F1Old_ = F1;
	R_   = R2;
	d_ = 2.0*R_;
	Rdt_ = Rdt2;
	
	
	if (R_ < R0_)
	  {
	    pG_ = pG0_*pow(R0_/R_,3.0);
	  }
	else
	  {
	    pG_ = pG0_*pow(R0_/R_,3.0*1.4);
	  }
	pB_ = pG_ + pV;
    }

 /* ======================================================= */    
 /* End of solving the Rayleigh-Plesset */


        if (onBoundary() && td.keepParticle)
        {
            if (isA<processorPolyPatch>(pbMesh[patch(face())]))
            {
                td.switchProcessor = true;
            }
        }
     //   #include "LPTtoVOF.H"  
    }

    return td.keepParticle;
}



/* Rayleigh-Plesset: initial equilibrium radius */

bool Foam::solidParticle::initialEquilibriumRadius
(
    trackingData& td
)
{
    td.switchProcessor = false;
    td.keepParticle = true;
    
    // remember which cell the parcel is in
    // since this will change if a face is hit
    label cellI = cell();
    
    cellPointWeight cpw(mesh_, position(), cellI, face());
    scalar pc  = td.pInterp().interpolate(cpw);
    
    
    /// initial equlibirium radius
    scalar p0 = td.cloud().p0();
    scalar pV = td.cloud().pSat();
    scalar sigma = td.cloud().sigma();
    scalar gamma = td.cloud().gamma();
    scalar R  = 0.5*d_;
    R_        = 0.5*d_;
    Rdt_      = 0.0; 
    F0Old_ = 0.0;
    F1Old_ = 0.0;

    
    // initial guess R0 equal to R
    scalar R0 = R;
    scalar R0new = R;
    scalar R0bar = R;
    scalar R0Old = R0;
    scalar error = 1;

     
    scalar a = (2.0*sigma)/(p0-pV);
    int maxIter = 2000;
    int iter    = 0;
    while ((error > 1.0E-06) && (iter < maxIter))
      {
	if (R < R0)
	  {
	    gamma = 1.4;
	  }
	else
	  {
	    gamma = 1.0;
	  }
	scalar  b = (pow(R,3.0*gamma)/(p0-pV))*(pV-((2.0*sigma)/R)-pc);
	scalar  f = pow(R0,3.0*gamma) + a*pow(R0,3.0*gamma-1.0) + b;
	scalar df = 3.0*gamma*pow(R0,3*gamma-1.0) + 
	  a*(3.0*gamma-1.0)*pow(R0,3.0*gamma-2.0);
	R0bar  = R0 - (f/df);
	if (R < R0bar)
	  {
	    gamma = 1.4;
	  }
	else
	  {
	    gamma = 1.0;
	  }
	b = (pow(R,3.0*gamma)/(p0-pV))*(pV-((2.0*sigma)/R)-pc);
	scalar fbar =  pow(R0bar,3.0*gamma) + a*pow(R0bar,3.0*gamma-1.0) + b;
	R0new  = R0 - (f*f)/(df*(f-fbar));
	R0new  = 0.9*R0 + 0.1*R0new;
	error = fabs((R0new-R0)/R0);
	R0     = R0new;
	iter++;
	
      }
     
    R0_  = R0;
    pG0_ = p0 + 2.0*sigma/R0_ - pV;
    if (R < R0)
      {
	gamma = 1.4;
      }
    else
      {
	gamma = 1.0;
      }
    pG_ = pG0_*pow(R0_/R_,3.0);
    pB_  = pG_ + pV;
    
    return true;
}




bool Foam::solidParticle::hitPatch
(
    const polyPatch&,
    trackingData&,
    const label,
    const scalar,
    const tetIndices&
)
{
    return false;
}


void Foam::solidParticle::hitProcessorPatch
(
    const processorPolyPatch&,
    trackingData& td
)
{
    td.switchProcessor = true;
}


void Foam::solidParticle::hitWallPatch
(
    const wallPolyPatch& wpp,
    trackingData& td,
    const tetIndices& tetIs
)
{
    vector nw = tetIs.faceTri(mesh_).normal();
    nw /= mag(nw);

    scalar Un = U_ & nw;
    vector Ut = U_ - Un*nw;

    if (Un > 0)
    {
        U_ -= (1.0 + td.cloud().e())*Un*nw;
    }

    U_ -= td.cloud().mu()*Ut;
}


void Foam::solidParticle::hitPatch
(
    const polyPatch&,
    trackingData& td
)
{
    td.keepParticle = false;
}


void Foam::solidParticle::transformProperties (const tensor& T)
{
    particle::transformProperties(T);
    U_ = transform(T, U_);
}


void Foam::solidParticle::transformProperties(const vector& separation)
{
    particle::transformProperties(separation);
}


Foam::scalar Foam::solidParticle::wallImpactDistance(const vector&) const
{
    return 0.5*d_;
}


// ************************************************************************* //
