/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 2011 OpenFOAM Foundation
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

\*---------------------------------------------------------------------------*/

#include "solidParticleCloud.H"
#include "fvMesh.H"
#include "volFields.H"
#include "interpolationCellPoint.H"

// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

Foam::solidParticleCloud::solidParticleCloud
(
    const fvMesh& mesh,
    const word& cloudName,
    bool readFields
)
:
    Cloud<solidParticle>(mesh, cloudName, false),
    mesh_(mesh),
    particleProperties_
    (
        IOobject
        (
            "particleProperties",
            mesh_.time().constant(),
            mesh_,
            IOobject::MUST_READ_IF_MODIFIED,
            IOobject::NO_WRITE
        )
    ),
    rhop_(dimensionedScalar(particleProperties_.lookup("rhop")).value()),
    e_(dimensionedScalar(particleProperties_.lookup("e")).value()),
    mu_(dimensionedScalar(particleProperties_.lookup("mu")).value()),
    smom_(mesh_.nCells(), vector::zero),  
    correctalphaW_(mesh_.nCells(), 0),    
    correctU_(mesh_.nCells(), vector::zero),   
    posP1_(dimensionedVector(particleProperties_.lookup("posP1")).value()),  
    dP1_(dimensionedScalar(particleProperties_.lookup("dP1")).value()),  
    UP1_(dimensionedVector(particleProperties_.lookup("UP1")).value()),   
    posP2_(dimensionedVector(particleProperties_.lookup("posP2")).value()),   
    dP2_(dimensionedScalar(particleProperties_.lookup("dP2")).value()),  
    UP2_(dimensionedVector(particleProperties_.lookup("UP2")).value()),  
    tInjStart_(dimensionedScalar(particleProperties_.lookup("tInjStart")).value()), 
    tInjEnd_(dimensionedScalar(particleProperties_.lookup("tInjEnd")).value()),
    p0_(dimensionedScalar(particleProperties_.lookup("p0")).value()),  
    pSat_(dimensionedScalar(particleProperties_.lookup("pSat")).value()),  
    gamma_(dimensionedScalar(particleProperties_.lookup("gamma")).value()),  
    sigma_(dimensionedScalar(particleProperties_.lookup("sigma")).value())  
{
    if (readFields)
    {
        solidParticle::readFields(*this);
    }
}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //


// Inject Function:
void Foam::solidParticleCloud::inject(solidParticle::trackingData &td)
{
	label cellI=1; 
    label tetFaceI=1; 
    label tetPtI=1; 
    mesh_.findCellFacePt(td.cloud().posP1_, cellI, tetFaceI, tetPtI);

    solidParticle* ptr1 = new solidParticle(mesh_, td.cloud().posP1_, cellI, tetFaceI, tetPtI,td.cloud().dP1_, td.cloud().UP1_);
    ptr1->initialEquilibriumRadius(td);
    Cloud<solidParticle>::addParticle(ptr1);
    
    mesh_.findCellFacePt(td.cloud().posP2_, cellI, tetFaceI, tetPtI);

    solidParticle* ptr2 = new solidParticle(mesh_, td.cloud().posP2_, cellI, tetFaceI, tetPtI, td.cloud().dP2_, td.cloud().UP2_);
    ptr2->initialEquilibriumRadius(td);
    Cloud<solidParticle>::addParticle(ptr2);
}

bool Foam::solidParticleCloud::hasWallImpactDistance() const
{
    return true;
}


void Foam::solidParticleCloud::move(const dimensionedVector& g)
{
    const volScalarField& rho = mesh_.lookupObject<const volScalarField>("rho");
    const volVectorField& U = mesh_.lookupObject<const volVectorField>("U");
    const volScalarField& nu = mesh_.lookupObject<const volScalarField>("nu");
    const volScalarField& alphaW = mesh_.lookupObject<const volScalarField>("alpha.water");  
    const volScalarField& p = mesh_.lookupObject<const volScalarField>("p"); 

    interpolationCellPoint<scalar> rhoInterp(rho);
    interpolationCellPoint<vector> UInterp(U);
    interpolationCellPoint<scalar> nuInterp(nu);
    interpolationCell<scalar> alphaWInterp(alphaW);    
    interpolationCellPoint<scalar> pInterp(p);
    
    smom_ = vector::zero;  
    correctalphaW_=0;  
    correctU_=vector::zero; 

    solidParticle::trackingData
        td(*this, rhoInterp, UInterp, nuInterp, alphaWInterp, pInterp, g.value());

    Cloud<solidParticle>::move(td, mesh_.time().deltaTValue());
    
    // Inject
    if(mesh_.time().value()> td.cloud().tInjStart_ && mesh_.time().value()< td.cloud().tInjEnd_)
    {
       this->inject(td);
    }
}


void Foam::solidParticleCloud::checkCo()
{
	if ((*this).size() < 2)
	{
		return;
	}
		
		Cloud<solidParticle>::iterator secondP = (*this).begin();
		++secondP;
		Cloud<solidParticle>::iterator p1 = secondP;
		while (p1 != (*this).end())
		{
			Cloud<solidParticle>::iterator p2 = (*this).begin();
			while (p2 != p1)
			{
				bool collision=false;
				#include "myCM.H"
				if (collision)
				{
					Info<< "Velocities before collision: "<< p1().U() <<p2().U()<<endl;
					
					vector p = p2().position() - p1().position();
					vector n12=p/mag(p);
					scalar v1n=p1().U() & n12;
					vector v1t=p1().U() - v1n*n12 ;
					scalar v2n=p2().U() & n12;
					vector v2t=p2().U() - v2n*n12 ;
					scalar COR=0.8;
					scalar m1 = rhop()*constant::mathematical::pi / 6.0*pow(p1().d(),3);
					scalar m2 = rhop()*constant::mathematical::pi / 6.0*pow(p2().d(),3);
					scalar mrn = m1*v1n + m2*v2n;
					scalar vnRel = v1n - v2n;
					p1().U_ = v1t+ ((mrn - COR* m2*vnRel)/(m1+m2))*n12;
					p2().U_ = v2t+ ((mrn + COR* m1*vnRel)/(m1+m2))*n12;
					
					Info<< "new velocities "<< p1().U() << " and "<<p2().U()<<endl;

					}
					++p2;
			} // end - inner loop
					++p1;
		} // end - outer loop
}


// ************************************************************************* //
